#'@rdname CST_Analogs
#'@title Downscaling using Analogs based on large scale fields.
#' 
#'@author J. Ramon, \email{jaume.ramon@bsc.es}
#'
#'@description This function performs a downscaling using Analogs. To compute 
#'the analogs given a coarse-scale field, the function looks for days with 
#'similar conditions in the historical observations. The coarse scale and 
#'observation data can be either global or regional. In the latter case, the 
#'region is defined by the user. In principle, the coarse and observation data 
#'should be of the same variable, although different variables can also be admitted. 
#'The analogs function will find the N best analogs based in Minimum Euclidean 
#'distance.
#'
#'The search of analogs must be done in the longest dataset posible, but might 
#'require high-memory computational resources. This is important since it is 
#'necessary to have a good representation of the possible states of the field in 
#'the past, and therefore, to get better analogs. The function can also look for 
#'analogs within a window of D days, but is the user who has to define that window. 
#'Otherwise, the function will look for analogs in the whole dataset. This function 
#'is intended to downscale climate prediction data (i.e., sub-seasonal, seasonal 
#'and decadal predictions) but can admit climate projections or reanalyses. It does 
#'not have constrains of specific region or variables to downscale.
#'@param exp an 's2dv' object with named dimensions containing the experimental field on
#'the coarse scale for which the downscaling is aimed. The object must have, at least,
#'the dimensions latitude, longitude, start date and time. The object is expected to be 
#'already subset for the desired region. Data can be in one or two integrated regions, e.g.,
#'crossing the Greenwich meridian. To get the correct results in the latter case,
#'the borders of the region should be specified in the parameter 'region'. See parameter
#''region'.
#'@param obs an 's2dv' object with named dimensions containing the observational field. 
#'The object must have, at least, the dimensions latitude, longitude and start date. 
#'The object is expected to be already subset for the desired region. 
#'@param nanalogs an integer indicating the number of analogs to be searched
#'@param fun_analog a function to be applied over the found analogs. Only these options
#'are valid: "mean", "wmean", "max", "min", "median" or NULL. If set to NULL (default), 
#'the function returns the found analogs.
#'@param lat_dim a character vector indicating the latitude dimension name in the element
#''data' in exp and obs. Default set to "lat".
#'@param lon_dim a character vector indicating the longitude dimension name in the element
#''data' in exp and obs. Default set to "lon".
#'@param sdate_dim a character vector indicating the start date dimension name in the
#'element 'data' in exp and obs. Default set to "sdate".
#'@param time_dim a character vector indicating the time dimension name in the element
#''data' in exp and obs. Default set to "time". 
#'@param region a numeric vector indicating the borders of the region defined in exp.
#'It consists of four elements in this order: lonmin, lonmax, latmin, latmax. lonmin refers
#'to the left border, while lonmax refers to the right border. latmin indicates the lower
#'border, whereas latmax indicates the upper border. If set to NULL (default), the function
#'takes the first and last elements of the latitudes and longitudes.
#'@param ncores an integer indicating the number of cores to use in parallel computation.
#'
#'@return An 's2dv' object. The element 'data' contains the dowscaled field, 'lat' the
#'downscaled latitudes, and 'lon' the downscaled longitudes. If fun_analog is set to NULL
#'(default), the output array in 'data' also contains the dimension 'analog' with the best 
#'analog days.
#'@examples
#'exp <- rnorm(15000)
#'dim(exp) <- c(member = 5, lat = 4, lon = 5, sdate = 5, time  = 30)
#'exp_lons <- 1:5
#'exp_lats <- 1:4
#'obs <- rnorm(27000)
#'dim(obs) <- c(lat = 12, lon = 15, sdate = 5, time  = 30)
#'obs_lons <- seq(0,6, 6/14)
#'obs_lats <- seq(0,6, 6/11)
#'exp <- s2dv_cube(data = exp, lat = exp_lats, lon = exp_lons)
#'obs <- s2dv_cube(data = obs, lat = obs_lats, lon = obs_lons)
#'downscaled_field <- CST_Analogs(exp = exp, obs = obs, grid_exp = 'r360x180')
#'@export
CST_Analogs <- function(exp, obs, grid_exp, nanalogs = 3, fun_analog = NULL, lat_dim = "lat", lon_dim = "lon",
                        sdate_dim = "sdate", time_dim = "time", region = NULL, ncores = 1) {

  # input exp and obs must be s2dv_cube objects
  if (!inherits(exp,'s2dv_cube')) {
    stop("Parameter 'exp' must be of the class 's2dv_cube'")
  }

  # input exp and obs must be s2dv_cube objects
  if (!inherits(obs,'s2dv_cube')) {
    stop("Parameter 'obs' must be of the class 's2dv_cube'")
  }

  res <- Analogs(exp = exp$data, obs = obs$data, exp_lats = exp[[lat_dim]], exp_lons = exp[[lon_dim]],
                 obs_lats = obs[[lat_dim]], obs_lons = obs[[lon_dim]], grid_exp = grid_exp, 
                 nanalogs = nanalogs, fun_analog = fun_analog, lat_dim = lat_dim, lon_dim = lon_dim, 
                 sdate_dim = sdate_dim, time_dim = time_dim, region = region, ncores = ncores)

  res_s2dv <- suppressWarnings(s2dv_cube(data = res$data, lon = res$lon, lat = res$lat))

  return(res_s2dv)

}

#'@rdname Analogs
#'@title Downscaling using Analogs based on large scale fields.
#' 
#'@author J. Ramon, \email{jaume.ramon@bsc.es}
#'@author Ll. Lledó, \email{llorenc.lledo@ecmwf.int}
#'
#'@description This function performs a downscaling using Analogs. To compute 
#'the analogs given a coarse-scale field, the function looks for days with 
#'similar conditions in the historical observations. The coarse scale and 
#'observation data can be either global or regional. In the latter case, the 
#'region is defined by the user. In principle, the coarse and observation data 
#'should be of the same variable, although different variables can also be admitted. 
#'The analogs function will find the N best analogs based in Minimum Euclidean 
#'distance.
#'
#'The search of analogs must be done in the longest dataset posible, but might 
#'require high-memory computational resources. This is important since it is 
#'necessary to have a good representation of the possible states of the field in 
#'the past, and therefore, to get better analogs. The function can also look for 
#'analogs within a window of D days, but is the user who has to define that window. 
#'Otherwise, the function will look for analogs in the whole dataset. This function 
#'is intended to downscale climate prediction data (i.e., sub-seasonal, seasonal 
#'and decadal predictions) but can admit climate projections or reanalyses. It does 
#'not have constrains of specific region or variables to downscale.
#'@param exp an array with named dimensions containing the experimental field on the
#'coarse scale for which the downscaling is aimed. The object must have, at least,
#'the dimensions latitude, longitude, start date and time. The object is expected to be 
#'already subset for the desired region. Data can be in one or two integrated regions, e.g.,
#'crossing the Greenwich meridian. To get the correct results in the latter case,
#'the borders of the region should be specified in the parameter 'region'. See parameter
#''region'.
#'@param obs an array with named dimensions containing the observational field. The object 
#'must have, at least, the dimensions latitude, longitude and start date. The object is 
#'expected to be already subset for the desired region. 
#'@param exp_lats a numeric vector containing the latitude values in 'exp'. Latitudes must 
#'range from -90 to 90.
#'@param exp_lons a numeric vector containing the longitude values in 'exp'. Longitudes 
#'can range from -180 to 180 or from 0 to 360.
#'@param obs_lats a numeric vector containing the latitude values in 'obs'. Latitudes must
#'range from -90 to 90.
#'@param obs_lons a numeric vector containing the longitude values in 'obs'. Longitudes
#'can range from -180 to 180 or from 0 to 360.
#'@param grid_exp a character vector with a path to an example file of the exp data.
#'@param nanalogs an integer indicating the number of analogs to be searched
#'@param fun_analog a function to be applied over the found analogs. Only these options
#'are valid: "mean", "wmean", "max", "min", "median" or NULL. If set to NULL (default), 
#'the function returns the found analogs.
#'@param lat_dim a character vector indicating the latitude dimension name in the element
#''data' in exp and obs. Default set to "lat".
#'@param lon_dim a character vector indicating the longitude dimension name in the element
#''data' in exp and obs. Default set to "lon".
#'@param sdate_dim a character vector indicating the start date dimension name in the
#'element 'data' in exp and obs. Default set to "sdate".
#'@param time_dim a character vector indicating the time dimension name in the element
#''data' in exp and obs. Default set to "time". 
#'@param region a numeric vector indicating the borders of the region defined in exp.
#'It consists of four elements in this order: lonmin, lonmax, latmin, latmax. lonmin refers
#'to the left border, while lonmax refers to the right border. latmin indicates the lower
#'border, whereas latmax indicates the upper border. If set to NULL (default), the function
#'takes the first and last elements of the latitudes and longitudes.
#'@param ncores an integer indicating the number of cores to use in parallel computation.
#'@import multiApply
#'@import CSTools
#'@importFrom s2dv InsertDim CDORemap
#'@importFrom FNN get.knnx
#'
#'@seealso \code{\link[s2dverification]{CDORemap}}
#'
#'@return A list of three elements. 'data' contains the dowscaled field, 'lat' the
#'downscaled latitudes, and 'lon' the downscaled longitudes. If fun_analog is set to NULL
#'(default), the output array in 'data' also contains the dimension 'analog' with the best 
#'analog days.
#'@examples
#'exp <- rnorm(15000)
#'dim(exp) <- c(member = 5, lat = 4, lon = 5, sdate = 5, time  = 30)
#'exp_lons <- 1:5
#'exp_lats <- 1:4
#'obs <- rnorm(27000)
#'dim(obs) <- c(lat = 12, lon = 15, sdate = 5, time  = 30)
#'obs_lons <- seq(0,6, 6/14)
#'obs_lats <- seq(0,6, 6/11)
#'downscaled_field <- Analogs(exp = exp, obs = obs, exp_lats = exp_lats, exp_lons = exp_lons, 
#'obs_lats = obs_lats, obs_lons = obs_lons, grid_exp = 'r360x180')
#'@export
Analogs <- function(exp, obs, exp_lats, exp_lons, obs_lats, obs_lons, grid_exp, nanalogs = 3, 
                    fun_analog = NULL, lat_dim = "lat", lon_dim = "lon", sdate_dim = "sdate", 
                    time_dim = "time", region = NULL, ncores = 1) {
  #-----------------------------------
  # Checkings
  #-----------------------------------
  if (!inherits(grid_exp, 'character')) {
    stop("Parameter 'grid_exp' must be of class 'character'. It can be either a path ",
         "to another NetCDF file which to read the target grid from (a single grid must be ",
         "defined in such file) or a character vector indicating the coarse grid to ",
         "be passed to CDO, and it must be a grid recognised by CDO or a NetCDF file.")
  }

  if (!inherits(nanalogs, 'numeric')) {
    stop("Parameter 'nanalogs' must be of the class 'character'")
  }

  if (!inherits(lat_dim, 'character')) {
    stop("Parameter 'lat_dim' must be of the class 'character'")
  }

  if (!inherits(lon_dim, 'character')) {
    stop("Parameter 'lon_dim' must be of the class 'character'")
  }

  if (!inherits(sdate_dim, 'character')) {
    stop("Parameter 'sdate_dim' must be of the class 'character'")
  }

  if (!inherits(time_dim, 'character')) {
    stop("Parameter 'time_dim' must be of the class 'character'")
  }

  # Do not allow synonims for lat (latitude), lon (longitude) and time (sdate) dimension names 
  if (is.na(match(lon_dim, names(dim(exp)))) | is.na(match(lon_dim, names(dim(obs))))) {
    stop("Missing longitude dimension in 'exp' and/or 'obs', or does not match the parameter ",
         "'lon_dim'")
  } 

  if (is.na(match(lat_dim, names(dim(exp)))) | is.na(match(lat_dim, names(dim(obs))))) {
    stop("Missing latitude dimension in 'exp' and/or 'obs', or does not match the parameter ",
         "'lat_dim'")
  }

  if (is.na(match(sdate_dim, names(dim(exp)))) | is.na(match(sdate_dim, names(dim(obs))))) {
    stop("Missing start date dimension in 'exp' and/or 'obs', or does not match the parameter ",
         "'sdate_dim'")
  }

  if (is.na(match(time_dim, names(dim(exp)))) | is.na(match(time_dim, names(dim(obs))))) {
    stop("Missing time dimension in 'exp' and/or 'obs', or does not match the parameter ",
         "'time_dim'")
  }

  # Ensure we have enough data to interpolate from high-res to coarse grid
  if ((obs_lats[1] > exp_lats[1]) | (obs_lats[length(obs_lats)] < exp_lats[length(exp_lats)]) |
      (obs_lons[1] > exp_lons[1]) | (obs_lons[length(obs_lons)] < exp_lons[length(exp_lons)])) {

    stop("There are not enough data in 'obs'. Please to add more latitudes or ",
         "longitudes.")
  }

  # Select a function to apply to the analogs selected for a given observation
  if (!is.null(fun_analog)) {
    stopifnot(fun_analog %in% c("mean", "wmean", "max", "min", "median"))
  }

  # Create window if user does not have it in obs
  if ( !("window" %in% names(dim(obs))) ) {
    nsdates <- dim(obs)[names(dim(obs)) == sdate_dim]
    ntimes <- dim(obs)[names(dim(obs)) == time_dim]
    window <- Apply(list(obs), target_dims = list(c(time_dim, sdate_dim)), 
                    fun = as.vector, output_dims = 'window')$output1
    obs <- InsertDim(obs, posdim = 1, lendim = nsdates * ntimes, name = "window")
    obs <- Apply(list(obs, window), target_dims = 'window', 
                 fun = function(x,y) x <- y)$output1
  }

  #-----------------------------------
  # Interpolate high-res observations to the coarse grid
  #-----------------------------------
  if (is.null(region)) {
    warning("The borders of the downscaling region have not been provided. Assuming the four borders of the ",
            "downscaling region are defined by the first and last elements of the parametres 'exp_lats' and 'exp_lons'.")
    region <- c(exp_lons[1], exp_lons[length(exp_lons)], exp_lats[1], exp_lats[length(exp_lats)])
  }

  obs_interpolated <- Interpolation(exp = obs, lats = obs_lats, lons = obs_lons, target_grid = grid_exp, 
                                    lat_dim = lat_dim, lon_dim = lon_dim, method_remap = "conservative", region = region)
  # If after interpolating 'obs' data the coordinates do not match, the exp data is interpolated to
  # the same grid to force the matching
  if (!.check_coords(lat1 = obs_interpolated$lat, lat2 = exp_lats, lon1 = obs_interpolated$lon, lon2 = exp_lons)) {
    exp_interpolated <- Interpolation(exp = exp, lats = exp_lats, lons = exp_lons, target_grid = grid_exp, 
                                      lat_dim = lat_dim, lon_dim = lon_dim, method_remap = "conservative", 
                                      region = region)$data
  } else {
    exp_interpolated <- exp
  }

  #-----------------------------------
  # Reshape train and test
  #-----------------------------------
  res <- Apply(list(obs_interpolated$data, exp_interpolated, obs), 
               target_dims = list(c("window", sdate_dim, time_dim, lat_dim, lon_dim), 
               c(sdate_dim, time_dim, lat_dim, lon_dim), c("window", sdate_dim, time_dim, lat_dim, lon_dim)), 
               fun = function(tr, te, ob) .analogs(train = tr, test = te, obs_hres = ob, k = nanalogs, 
               fun_analog = fun_analog), ncores = ncores)$output1

  #test <- exp_interpolated[1,1,,1,,,]
  #train <- obs_interpolated$data[,1,1,,,,]
  #obs_hres <- obs[,1,1,,,,]
  #test <- aperm(test,c(2,1,3,4))
  #train <- aperm(train,c(1,4,2,3,5))
  #obs_hres <- aperm(obs_hres,c(1,4,2,3,5))  

  res <- list(data = res, lon = obs_lons, lat = obs_lats)
  
  return(res)
}

# For each element in test, find the indices of the k nearest neigbhors in train
.analogs <- function(train, test, obs_hres, k, fun_analog) {
  # train, test, and obs_hres  dim: 4 dimensions sdate, time, lat and lon (in this order)

  # Number of lats/lons of the high-resolution data
  space_dims_hres <- dim(obs_hres)[c(4,5)]
  nsdates <-  dim(train)[2]
  ntimes <- dim(train)[3]

  # Reformat train and test as an array with (time, points)
  train <- apply(train, c(1,2,3), as.vector); names(dim(train))[1] <- "space"  
  test <- apply(test, c(1,2), as.vector); names(dim(test))[1] <- "space"
  obs_hres <- apply(obs_hres, c(1,2,3), as.vector); names(dim(obs_hres))[1] <- "space"
  
  # Here we do cross-validation: the start date considered is removed from the training
  analogs_arr <- array(NA, dim = c(analogs = k, space_dims_hres, ntimes, nsdates))
  for (s in 1:nsdates) {
    for (t in 1:ntimes) {
      tr <- apply(train[ , , -s, t], 1, as.vector); names(dim(tr))[1] <- "time"
      te <- test[ , s, t]
      ob <- apply(obs_hres[ , , -s, t], 1, as.vector); names(dim(ob))[1] <- "time"

      # Find NA's and remove them 
      # Assume that NA's are only in the spatial dimension, not in the temporal
      idx_na_tr <- is.na(tr[1 , ])
      idx_na_te <- is.na(te)
      idx_na <- idx_na_tr | idx_na_te
      tr_wo_na <- tr[ , !idx_na]
      te_wo_na <- te[!idx_na]
      te_wo_na <- InsertDim(data = te_wo_na, posdim = 1, lendim = 1, name = 'time') 
      names(dim(te_wo_na))[2] <- "space"
      
      knn.ind <- get.knnx(tr_wo_na, te_wo_na, k)

      dist <- knn.ind$nn.dist
      idx <- knn.ind$nn.index

      analogs <- ob[ idx, ]      
      dim(analogs) <- c(analogs = k, space_dims_hres)
     
      analogs_arr[ , , , t, s] <- analogs
    }    
  }

  # Apply functions to analogs 
  if (!is.null(fun_analog)) {
    if (fun_analog == "wmean") {
      weight <- 1 / dist
      analogs_arr <- apply(analogs_arr, c(2,3,4,5), function(x) weighted.mean(x, weight))
    } else {
      analogs_arr <- apply(analogs_arr, c(2,3,4,5), fun_analog)
    }
  }

  return(analogs_arr)
}
  

